"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.promiseSome = void 0;
// eslint-disable-next-line @typescript-eslint/no-empty-function
const infinitePromise = new Promise(() => { });
/**
 * A function that simulates the Array.some behaviour but on an array of Promises.
 * Returns true if at least one promise returns true.
 * Returns false if all promises return false or error.
 *
 * @remarks
 *
 * Predicates provided as input must be implemented considering
 * the following points:
 * 1. if they throw an error, it won't be propagated;
 * 2. throwing an error should be logically equivalent to returning false.
 */
async function promiseSome(predicates) {
    // These promises will only finish when their predicate returns true
    const infinitePredicates = predicates.map(async (predicate) => predicate.then(async (value) => value ? true : infinitePromise, async () => infinitePromise));
    // Returns after all predicates are resolved
    const finalPromise = Promise.allSettled(predicates).then((results) => results.some((result) => result.status === 'fulfilled' && result.value));
    // Either one of the infinitePredicates will return true,
    // or finalPromise will return the result if none of them did or finalPromise was faster
    return Promise.race([...infinitePredicates, finalPromise]);
}
exports.promiseSome = promiseSome;
//# sourceMappingURL=PromiseUtil.js.map