"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.filterHandlers = exports.findHandler = exports.createAggregateError = void 0;
const BadRequestHttpError_1 = require("../errors/BadRequestHttpError");
const ErrorUtil_1 = require("../errors/ErrorUtil");
const HttpError_1 = require("../errors/HttpError");
const InternalServerError_1 = require("../errors/InternalServerError");
/**
 * Combines a list of errors into a single HttpErrors.
 * Status code depends on the input errors. If they all share the same status code that code will be re-used.
 * If they are all within the 4xx range, 400 will be used, otherwise 500.
 *
 * @param errors - Errors to combine.
 * @param messagePrefix - Prefix for the aggregate error message. Will be followed with an array of all the messages.
 */
function createAggregateError(errors, messagePrefix = 'No handler supports the given input:') {
    const httpErrors = errors.map((error) => HttpError_1.HttpError.isInstance(error) ? error : new InternalServerError_1.InternalServerError(ErrorUtil_1.createErrorMessage(error)));
    const joined = httpErrors.map((error) => error.message).join(', ');
    const message = `${messagePrefix} [${joined}]`;
    // Check if all errors have the same status code
    if (httpErrors.length > 0 && httpErrors.every((error) => error.statusCode === httpErrors[0].statusCode)) {
        return new HttpError_1.HttpError(httpErrors[0].statusCode, httpErrors[0].name, message);
    }
    // Find the error range (4xx or 5xx)
    if (httpErrors.some((error) => error.statusCode >= 500)) {
        return new InternalServerError_1.InternalServerError(message);
    }
    return new BadRequestHttpError_1.BadRequestHttpError(message);
}
exports.createAggregateError = createAggregateError;
/**
 * Finds a handler that can handle the given input data.
 * Otherwise an error gets thrown.
 *
 * @param handlers - List of handlers to search in.
 * @param input - The input data.
 *
 * @returns A promise resolving to a handler that supports the data or otherwise rejecting.
 */
async function findHandler(handlers, input) {
    const errors = [];
    for (const handler of handlers) {
        try {
            await handler.canHandle(input);
            return handler;
        }
        catch (error) {
            if (ErrorUtil_1.isError(error)) {
                errors.push(error);
            }
            else {
                errors.push(new Error(ErrorUtil_1.createErrorMessage(error)));
            }
        }
    }
    throw createAggregateError(errors);
}
exports.findHandler = findHandler;
/**
 * Filters a list of handlers to only keep those that can handle the input.
 * Will error if no matching handlers are found.
 *
 * @param handlers - Handlers to filter.
 * @param input - Input that needs to be supported.
 */
async function filterHandlers(handlers, input) {
    const results = await Promise.allSettled(handlers.map(async (handler) => {
        await handler.canHandle(input);
        return handler;
    }));
    const matches = results.filter(({ status }) => status === 'fulfilled')
        .map((result) => result.value);
    if (matches.length > 0) {
        return matches;
    }
    // Generate error in case no matches were found
    const errors = results.map((result) => result.reason);
    throw createAggregateError(errors);
}
exports.filterHandlers = filterHandlers;
//# sourceMappingURL=HandlerUtil.js.map