import { AsyncHandler } from './AsyncHandler';
declare type ThenArg<T> = T extends PromiseLike<infer U> ? U : T;
declare type InType<T extends AsyncHandler<any, any>> = Parameters<T['handle']>[0];
declare type OutType<T extends AsyncHandler<any, any>> = ThenArg<ReturnType<T['handle']>>;
/**
 * Utility handler that allows combining the results of multiple handlers into one.
 * Will run all the handlers and then call the abstract `combine` function with the results,
 * which should return the output of the class.
 *
 * If `requireAll` is true, the handler will fail if any of the handlers do not support the input.
 * If `requireAll` is false, only the handlers that support the input will be called,
 * only if all handlers reject the input will this handler reject as well.
 * With `requireAll` set to false, the length of the input array
 * for the `combine` function is variable (but always at least 1).
 */
export declare abstract class UnionHandler<T extends AsyncHandler<any, any>> extends AsyncHandler<InType<T>, OutType<T>> {
    protected readonly handlers: T[];
    private readonly requireAll;
    protected constructor(handlers: T[], requireAll?: boolean);
    canHandle(input: InType<T>): Promise<void>;
    handle(input: InType<T>): Promise<OutType<T>>;
    handleSafe(input: InType<T>): Promise<OutType<T>>;
    /**
     * Checks if all handlers can handle the input.
     * If not, throw an error based on the errors of the failed handlers.
     */
    private allCanHandle;
    /**
     * Combine the results of the handlers into a single output.
     */
    protected abstract combine(results: OutType<T>[]): Promise<OutType<T>>;
}
export {};
