"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.UnionHandler = void 0;
const AsyncHandler_1 = require("./AsyncHandler");
const HandlerUtil_1 = require("./HandlerUtil");
/**
 * Utility handler that allows combining the results of multiple handlers into one.
 * Will run all the handlers and then call the abstract `combine` function with the results,
 * which should return the output of the class.
 *
 * If `requireAll` is true, the handler will fail if any of the handlers do not support the input.
 * If `requireAll` is false, only the handlers that support the input will be called,
 * only if all handlers reject the input will this handler reject as well.
 * With `requireAll` set to false, the length of the input array
 * for the `combine` function is variable (but always at least 1).
 */
class UnionHandler extends AsyncHandler_1.AsyncHandler {
    constructor(handlers, requireAll = false) {
        super();
        this.handlers = handlers;
        this.requireAll = requireAll;
    }
    async canHandle(input) {
        if (this.requireAll) {
            await this.allCanHandle(input);
        }
        else {
            // This will error if no handler supports the input
            await HandlerUtil_1.findHandler(this.handlers, input);
        }
    }
    async handle(input) {
        let handlers;
        if (this.requireAll) {
            // Handlers were already checked in canHandle
            // eslint-disable-next-line prefer-destructuring
            handlers = this.handlers;
        }
        else {
            handlers = await HandlerUtil_1.filterHandlers(this.handlers, input);
        }
        const results = await Promise.all(handlers.map(async (handler) => handler.handle(input)));
        return this.combine(results);
    }
    async handleSafe(input) {
        let handlers;
        if (this.requireAll) {
            await this.allCanHandle(input);
            // eslint-disable-next-line prefer-destructuring
            handlers = this.handlers;
        }
        else {
            // This will error if no handler supports the input
            handlers = await HandlerUtil_1.filterHandlers(this.handlers, input);
        }
        const results = await Promise.all(handlers.map(async (handler) => handler.handle(input)));
        return this.combine(results);
    }
    /**
     * Checks if all handlers can handle the input.
     * If not, throw an error based on the errors of the failed handlers.
     */
    async allCanHandle(input) {
        const results = await Promise.allSettled(this.handlers.map(async (handler) => {
            await handler.canHandle(input);
            return handler;
        }));
        if (results.some(({ status }) => status === 'rejected')) {
            const errors = results.map((result) => result.reason);
            throw HandlerUtil_1.createAggregateError(errors);
        }
    }
}
exports.UnionHandler = UnionHandler;
//# sourceMappingURL=UnionHandler.js.map