"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const solid_client_authn_core_1 = require("@inrupt/solid-client-authn-core");
const fetchFactory_1 = require("../../../authenticatedFetch/fetchFactory");
function validateOptions(oidcLoginOptions) {
    return (oidcLoginOptions.refreshToken !== undefined &&
        oidcLoginOptions.client.clientId !== undefined);
}
async function refreshAccess(refreshOptions, dpop) {
    var _a, _b;
    let authFetch;
    let tokens;
    try {
        if (dpop) {
            const dpopKeys = await solid_client_authn_core_1.generateDpopKeyPair();
            dpopKeys.publicKey.alg = "ES256";
            tokens = await refreshOptions.tokenRefresher.refresh(refreshOptions.sessionId, refreshOptions.refreshToken, dpopKeys, refreshOptions.onNewRefreshToken);
            const rotatedRefreshOptions = {
                ...refreshOptions,
                refreshToken: (_a = tokens.refresh_token) !== null && _a !== void 0 ? _a : refreshOptions.refreshToken,
            };
            authFetch = await fetchFactory_1.buildDpopFetch(tokens.access_token, dpopKeys, rotatedRefreshOptions);
        }
        else {
            tokens = await refreshOptions.tokenRefresher.refresh(refreshOptions.sessionId, refreshOptions.refreshToken, undefined, refreshOptions.onNewRefreshToken);
            const rotatedRefreshOptions = {
                ...refreshOptions,
                refreshToken: (_b = tokens.refresh_token) !== null && _b !== void 0 ? _b : refreshOptions.refreshToken,
            };
            authFetch = fetchFactory_1.buildBearerFetch(tokens.access_token, rotatedRefreshOptions);
        }
    }
    catch (e) {
        throw new Error(`Invalid refresh credentials: ${e.toString()}`);
    }
    return Object.assign(tokens, {
        fetch: authFetch,
    });
}
class RefreshTokenOidcHandler {
    constructor(tokenRefresher, storageUtility) {
        this.tokenRefresher = tokenRefresher;
        this.storageUtility = storageUtility;
    }
    async canHandle(oidcLoginOptions) {
        return validateOptions(oidcLoginOptions);
    }
    async handle(oidcLoginOptions) {
        var _a;
        if (!(await this.canHandle(oidcLoginOptions))) {
            throw new Error(`RefreshTokenOidcHandler cannot handle the provided options, missing one of 'refreshToken', 'clientId' in: ${JSON.stringify(oidcLoginOptions)}`);
        }
        const refreshOptions = {
            refreshToken: oidcLoginOptions.refreshToken,
            sessionId: oidcLoginOptions.sessionId,
            tokenRefresher: this.tokenRefresher,
            onNewRefreshToken: oidcLoginOptions.onNewRefreshToken,
        };
        await this.storageUtility.setForUser(oidcLoginOptions.sessionId, {
            issuer: oidcLoginOptions.issuer,
            dpop: oidcLoginOptions.dpop ? "true" : "false",
            clientId: oidcLoginOptions.client.clientId,
            clientSecret: oidcLoginOptions.client.clientSecret,
        });
        const accessInfo = await refreshAccess(refreshOptions, oidcLoginOptions.dpop);
        const sessionInfo = {
            isLoggedIn: true,
            sessionId: oidcLoginOptions.sessionId,
        };
        if (accessInfo.id_token === undefined) {
            throw new Error(`The Identity Provider [${oidcLoginOptions.issuer}] did not return an ID token on refresh, which prevents us from getting the user's WebID.`);
        }
        sessionInfo.webId = await await solid_client_authn_core_1.getWebidFromTokenPayload(accessInfo.id_token, oidcLoginOptions.issuerConfiguration.jwksUri, oidcLoginOptions.issuer, oidcLoginOptions.client.clientId);
        await solid_client_authn_core_1.saveSessionInfoToStorage(this.storageUtility, oidcLoginOptions.sessionId, undefined, undefined, "true", (_a = accessInfo.refresh_token) !== null && _a !== void 0 ? _a : refreshOptions.refreshToken);
        await this.storageUtility.setForUser(oidcLoginOptions.sessionId, {
            issuer: oidcLoginOptions.issuer,
            dpop: oidcLoginOptions.dpop ? "true" : "false",
            clientId: oidcLoginOptions.client.clientId,
        });
        if (oidcLoginOptions.client.clientSecret) {
            await this.storageUtility.setForUser(oidcLoginOptions.sessionId, {
                clientSecret: oidcLoginOptions.client.clientSecret,
            });
        }
        if (oidcLoginOptions.client.clientName) {
            await this.storageUtility.setForUser(oidcLoginOptions.sessionId, {
                clientName: oidcLoginOptions.client.clientName,
            });
        }
        return Object.assign(sessionInfo, {
            fetch: accessInfo.fetch,
        });
    }
}
exports.default = RefreshTokenOidcHandler;
//# sourceMappingURL=RefreshTokenOidcHandler.js.map