"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const solid_client_authn_core_1 = require("@inrupt/solid-client-authn-core");
const openid_client_1 = require("openid-client");
const IssuerConfigFetcher_1 = require("../IssuerConfigFetcher");
const ClientRegistrar_1 = require("../ClientRegistrar");
class TokenRefresher {
    constructor(storageUtility, issuerConfigFetcher, clientRegistrar) {
        this.storageUtility = storageUtility;
        this.issuerConfigFetcher = issuerConfigFetcher;
        this.clientRegistrar = clientRegistrar;
    }
    async refresh(sessionId, refreshToken, dpopKey, onNewRefreshToken) {
        const oidcContext = await solid_client_authn_core_1.loadOidcContextFromStorage(sessionId, this.storageUtility, this.issuerConfigFetcher);
        const issuer = new openid_client_1.Issuer(IssuerConfigFetcher_1.configToIssuerMetadata(oidcContext.issuerConfig));
        const clientInfo = await this.clientRegistrar.getClient({ sessionId }, oidcContext.issuerConfig);
        if (clientInfo.idTokenSignedResponseAlg === undefined) {
            clientInfo.idTokenSignedResponseAlg = ClientRegistrar_1.negotiateClientSigningAlg(oidcContext.issuerConfig, solid_client_authn_core_1.PREFERRED_SIGNING_ALG);
        }
        const client = new issuer.Client({
            client_id: clientInfo.clientId,
            client_secret: clientInfo.clientSecret,
            id_token_signed_response_alg: clientInfo.idTokenSignedResponseAlg,
        });
        if (refreshToken === undefined) {
            throw new Error(`Session [${sessionId}] has no refresh token to allow it to refresh its access token.`);
        }
        if (oidcContext.dpop && dpopKey === undefined) {
            throw new Error(`For session [${sessionId}], the key bound to the DPoP access token must be provided to refresh said access token.`);
        }
        const tokenSet = await client.refresh(refreshToken, {
            DPoP: dpopKey ? dpopKey.privateKey : undefined,
        });
        if (tokenSet.access_token === undefined) {
            throw new Error(`The Identity Provider [${issuer.metadata.issuer}] did not return an access token on refresh.`);
        }
        if (tokenSet.refresh_token !== undefined) {
            await this.storageUtility.setForUser(sessionId, {
                refreshToken: tokenSet.refresh_token,
            });
            if (typeof onNewRefreshToken === "function") {
                onNewRefreshToken(tokenSet.refresh_token);
            }
        }
        return tokenSet;
    }
}
exports.default = TokenRefresher;
//# sourceMappingURL=TokenRefresher.js.map