"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.verifySolidAccessToken = void 0;
const verify_1 = __importDefault(require("jose/jwt/verify"));
const ts_guards_1 = require("ts-guards");
const standard_object_1 = require("ts-guards/dist/standard-object");
const config_1 = require("../config");
const SolidAccessTokenGuard_1 = require("../guard/SolidAccessTokenGuard");
const Crypto_1 = require("../type/Crypto");
const decodeBase64UrlEncodedJson_1 = require("./decodeBase64UrlEncodedJson");
const retrieveAccessTokenIssuerKeySet_1 = require("./retrieveAccessTokenIssuerKeySet");
const retrieveWebidTrustedOidcIssuers_1 = require("./retrieveWebidTrustedOidcIssuers");
const verifyDpopProof_1 = require("./verifyDpopProof");
const verifySecureUriClaim_1 = require("./verifySecureUriClaim");
const verifySolidAccessTokenIssuer_1 = require("./verifySolidAccessTokenIssuer");
const verifySolidAccessTokenRequiredClaims_1 = require("./verifySolidAccessTokenRequiredClaims");
/**
 * Verify Solid access token
 *
 * @param authorization The authorization header and optional key and trusted issuer retrieval functions
 * @param dpopOptions The DPoP proof header and associated verifiable claims
 * @returns Access token payload
 */
async function verifySolidAccessToken(authorization, dpopOptions) {
    // Extract access token value from authorization header
    const accessTokenValue = authorization.header.replace(/^(DPoP|Bearer) /, "");
    // Decode Solid access token payload
    const accessTokenPayload = (0, decodeBase64UrlEncodedJson_1.decodeBase64UrlEncodedJson)(accessTokenValue.split(".")[1]);
    // Verify the Solid access token includes all required claims
    (0, verifySolidAccessTokenRequiredClaims_1.verifySolidAccessTokenRequiredClaims)(accessTokenPayload);
    // Check WebID claim is a secure URI
    (0, verifySecureUriClaim_1.verifySecureUriClaim)(accessTokenPayload.webid, "webid");
    // Check Issuer claim is a secure URI
    (0, verifySecureUriClaim_1.verifySecureUriClaim)(accessTokenPayload.iss, "iss");
    // Retrieve the issuers listed in the WebID
    const issuers = await (0, retrieveWebidTrustedOidcIssuers_1.retrieveWebidTrustedOidcIssuers)(accessTokenPayload.webid, authorization.issuers);
    // Check the issuer claim matches one of the WebID's trusted issuers
    (0, verifySolidAccessTokenIssuer_1.verifySolidAccessTokenIssuer)(issuers, accessTokenPayload.iss);
    /**
     * Verify Access Token
     * - Signature of Access Token JWT/JWS matches a key in the remote jwks
     * - Access Token max age 1 day
     * - Claims:
     *    - audience 'aud' is solid
     *    - algorithm 'alg' is an asymetric cryptographic algorithm
     *    - expiration 'exp' is not in the past
     *    - 'iat' is not in the future
     */
    const { payload, protectedHeader } = await (0, verify_1.default)(accessTokenValue, await (0, retrieveAccessTokenIssuerKeySet_1.retrieveAccessTokenIssuerKeySet)(accessTokenPayload.iss, authorization.keySet), {
        audience: "solid",
        algorithms: Array.from(Crypto_1.asymetricCryptographicAlgorithm),
        maxTokenAge: `${config_1.maxAccessTokenAgeInSeconds}s`,
        clockTolerance: `${config_1.clockToleranceInSeconds}s`,
    });
    // Get JWT value for either DPoP or Bearer tokens
    const accessToken = {
        header: protectedHeader,
        payload,
        signature: accessTokenValue.split(".")[2],
    };
    (0, SolidAccessTokenGuard_1.isSolidAccessToken)(accessToken);
    if (authorization.header.startsWith("DPoP ") ||
        (0, standard_object_1.isObjectPropertyOf)(accessToken.payload, "cnf")) {
        try {
            ts_guards_1.asserts.isNotNullOrUndefined(dpopOptions);
        }
        catch (_) {
            throw new Error("SolidIdentityDPoPError DPoP options missing for DPoP bound access token verification");
        }
        await (0, verifyDpopProof_1.verifyDpopProof)(dpopOptions.header, accessToken, accessTokenValue, dpopOptions.method, dpopOptions.url, dpopOptions.isDuplicateJTI);
    }
    return accessToken.payload;
}
exports.verifySolidAccessToken = verifySolidAccessToken;
//# sourceMappingURL=verifySolidAccessToken.js.map