"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Generator = void 0;
const ComponentsManagerBuilder_1 = require("componentsjs/lib/loading/ComponentsManagerBuilder");
const PrefetchedDocumentLoader_1 = require("componentsjs/lib/rdf/PrefetchedDocumentLoader");
const jsonld_context_parser_1 = require("jsonld-context-parser");
const ClassFinder_1 = require("../parse/ClassFinder");
const ClassIndexer_1 = require("../parse/ClassIndexer");
const ClassLoader_1 = require("../parse/ClassLoader");
const ConstructorLoader_1 = require("../parse/ConstructorLoader");
const PackageMetadataLoader_1 = require("../parse/PackageMetadataLoader");
const ParameterResolver_1 = require("../parse/ParameterResolver");
const ExternalModulesLoader_1 = require("../resolution/ExternalModulesLoader");
const ComponentConstructor_1 = require("../serialize/ComponentConstructor");
const ComponentSerializer_1 = require("../serialize/ComponentSerializer");
const ContextConstructor_1 = require("../serialize/ContextConstructor");
/**
 * Generates a components file by parsing a typescript file.
 */
class Generator {
    constructor(args) {
        this.resolutionContext = args.resolutionContext;
        this.pathDestination = args.pathDestination;
        this.fileExtension = args.fileExtension;
        this.ignoreClasses = args.ignoreClasses;
        this.typeScopedContexts = args.typeScopedContexts;
        this.logLevel = args.logLevel;
        this.prefix = args.prefix;
    }
    async generateComponents() {
        const logger = ComponentsManagerBuilder_1.ComponentsManagerBuilder.createLogger(this.logLevel);
        // Load package metadata
        const packageMetadata = await new PackageMetadataLoader_1.PackageMetadataLoader({ resolutionContext: this.resolutionContext,
            prefix: this.prefix })
            .load(this.pathDestination.packageRootDirectory);
        const classLoader = new ClassLoader_1.ClassLoader({ resolutionContext: this.resolutionContext, logger });
        const classFinder = new ClassFinder_1.ClassFinder({ classLoader });
        const classIndexer = new ClassIndexer_1.ClassIndexer({ classLoader, classFinder, ignoreClasses: this.ignoreClasses, logger });
        // Find all relevant classes
        const packageExports = await classFinder.getPackageExports(packageMetadata.name, packageMetadata.typesPath);
        const classAndInterfaceIndex = await classIndexer.createIndex(packageExports);
        // Load constructor data
        const constructorsUnresolved = new ConstructorLoader_1.ConstructorLoader().getConstructors(classAndInterfaceIndex);
        const constructors = await new ParameterResolver_1.ParameterResolver({ classLoader, ignoreClasses: this.ignoreClasses })
            .resolveAllConstructorParameters(constructorsUnresolved, classAndInterfaceIndex);
        // Load external components
        const externalModulesLoader = new ExternalModulesLoader_1.ExternalModulesLoader({
            pathDestination: this.pathDestination,
            packageMetadata,
            logger,
        });
        const externalPackages = externalModulesLoader.findExternalPackages(classAndInterfaceIndex, constructors);
        const externalComponents = await externalModulesLoader.loadExternalComponents(require, externalPackages);
        // Create components
        const contextConstructor = new ContextConstructor_1.ContextConstructor({
            packageMetadata,
            typeScopedContexts: this.typeScopedContexts,
        });
        const componentConstructor = new ComponentConstructor_1.ComponentConstructor({
            packageMetadata,
            contextConstructor,
            pathDestination: this.pathDestination,
            classAndInterfaceIndex,
            classConstructors: constructors,
            externalComponents,
            contextParser: new jsonld_context_parser_1.ContextParser({
                documentLoader: new PrefetchedDocumentLoader_1.PrefetchedDocumentLoader({
                    contexts: externalComponents.moduleState.contexts,
                    logger,
                }),
                skipValidation: true,
            }),
        });
        const components = await componentConstructor.constructComponents();
        const componentsIndex = await componentConstructor.constructComponentsIndex(components, this.fileExtension);
        // Serialize components
        const componentSerializer = new ComponentSerializer_1.ComponentSerializer({
            resolutionContext: this.resolutionContext,
            pathDestination: this.pathDestination,
            fileExtension: this.fileExtension,
            indentation: '  ',
        });
        await componentSerializer.serializeComponents(components);
        await componentSerializer.serializeComponentsIndex(componentsIndex);
        // Serialize context
        const context = contextConstructor.constructContext(components);
        await componentSerializer.serializeContext(context);
    }
}
exports.Generator = Generator;
//# sourceMappingURL=Generator.js.map