import type { ClassDeclaration, TSInterfaceDeclaration } from '@typescript-eslint/types/dist/ts-estree';
import type { AST, TSESTreeOptions } from '@typescript-eslint/typescript-estree';
import type { Logger } from 'winston';
import type { ResolutionContext } from '../resolution/ResolutionContext';
import type { ClassLoaded, ClassReference, ClassReferenceLoaded, GenericTypes, InterfaceLoaded } from './ClassIndex';
/**
 * Loads typescript classes from class references.
 */
export declare class ClassLoader {
    private readonly resolutionContext;
    private readonly logger;
    constructor(args: ClassLoaderArgs);
    /**
     * Find the super class of the given class.
     * Throws an error for super class definitions that could not be interpreted.
     * @param declaration A class declaration.
     * @param fileName The file name of the current class.
     */
    getSuperClassName(declaration: ClassDeclaration, fileName: string): string | undefined;
    /**
     * Find the super interfaces of the given interface.
     * Throws an error for interface definitions that could not be interpreted.
     * @param declaration An interface declaration.
     * @param fileName The file name of the current class.
     */
    getSuperInterfaceNames(declaration: TSInterfaceDeclaration, fileName: string): string[];
    /**
     * Find the interface names of the given class.
     * @param declaration A class declaration.
     * @param fileName The file name of the current class.
     */
    getClassInterfaceNames(declaration: ClassDeclaration, fileName: string): string[];
    /**
     * Load the referenced class, and obtain its full class declaration.
     * Classes can either be defined in this file (exported or not), or imported from another file.
     * @param classReference The reference to a class.
     * @param considerInterfaces If the class reference is allows to refer to an interface, as well as a class.
     */
    loadClassDeclaration<CI extends boolean>(classReference: ClassReference, considerInterfaces: CI): Promise<CI extends true ? (ClassLoaded | InterfaceLoaded) : ClassLoaded>;
    /**
     * Create a hash of generic types in the given class declaration.
     * @param classDeclaration A class or interface declaration.
     */
    collectGenericTypes(classDeclaration: ClassDeclaration | TSInterfaceDeclaration): GenericTypes;
    /**
     * Annotate the given loaded class or interface with a comment if it is present on the declaration.
     * @param classLoaded A loaded class or interface.
     */
    enhanceLoadedWithComment(classLoaded: ClassReferenceLoaded): ClassReferenceLoaded;
    /**
     * Load a class, and get all class elements from it.
     * @param packageName Package name we are importing from.
     * @param fileName A file path.
     */
    loadClassElements(packageName: string, fileName: string): Promise<ClassElements>;
    /**
     * Convert the given import path to an absolute file path, coupled with the module it is part of.
     * @param currentPackageName Package name we are importing from.
     * @param currentFilePath Absolute path to a file in which the import path occurs.
     * @param importPath Possibly relative path that is being imported.
     */
    importTargetToAbsolutePath(currentPackageName: string, currentFilePath: string, importPath: string): {
        packageName: string;
        fileName: string;
    };
    /**
     * Get all class elements in a file.
     * @param packageName Package name we are importing from.
     * @param fileName A file path.
     * @param ast The parsed file.
     */
    getClassElements(packageName: string, fileName: string, ast: AST<TSESTreeOptions>): ClassElements;
}
export interface ClassLoaderArgs {
    resolutionContext: ResolutionContext;
    logger: Logger;
}
/**
 * Holder for all available classes in a file.
 */
export interface ClassElements {
    exportedClasses: Record<string, ClassDeclaration>;
    exportedInterfaces: Record<string, TSInterfaceDeclaration>;
    exportedImportedElements: Record<string, ClassReference>;
    exportedImportedAll: {
        packageName: string;
        fileName: string;
    }[];
    exportedUnknowns: Record<string, string>;
    declaredClasses: Record<string, ClassDeclaration>;
    declaredInterfaces: Record<string, TSInterfaceDeclaration>;
    importedElements: Record<string, ClassReference>;
}
