"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ClassLoader = void 0;
const Path = require("path");
const typescript_estree_1 = require("@typescript-eslint/typescript-estree");
const CommentLoader_1 = require("./CommentLoader");
/**
 * Loads typescript classes from class references.
 */
class ClassLoader {
    constructor(args) {
        this.resolutionContext = args.resolutionContext;
        this.logger = args.logger;
    }
    /**
     * Find the super class of the given class.
     * Throws an error for super class definitions that could not be interpreted.
     * @param declaration A class declaration.
     * @param fileName The file name of the current class.
     */
    getSuperClassName(declaration, fileName) {
        if (!declaration.superClass) {
            return;
        }
        if (declaration.superClass.type === typescript_estree_1.AST_NODE_TYPES.Identifier) {
            // Extensions in the form of `class A extends B`
            return declaration.superClass.name;
        }
        if (declaration.superClass.type === typescript_estree_1.AST_NODE_TYPES.MemberExpression &&
            declaration.superClass.property.type === typescript_estree_1.AST_NODE_TYPES.Identifier &&
            declaration.superClass.object.type === typescript_estree_1.AST_NODE_TYPES.Identifier) {
            // Extensions in the form of `class A extends x.B`
            throw new Error(`Namespaced superclasses are currently not supported: ${fileName} on line ${declaration.superClass.loc.start.line} column ${declaration.superClass.loc.start.column}`);
        }
        throw new Error(`Could not interpret type of superclass in ${fileName} on line ${declaration.superClass.loc.start.line} column ${declaration.superClass.loc.start.column}`);
    }
    /**
     * Find the super interfaces of the given interface.
     * Throws an error for interface definitions that could not be interpreted.
     * @param declaration An interface declaration.
     * @param fileName The file name of the current class.
     */
    getSuperInterfaceNames(declaration, fileName) {
        return (declaration.extends || [])
            // eslint-disable-next-line array-callback-return
            .map(extendsExpression => {
            if (extendsExpression.type === typescript_estree_1.AST_NODE_TYPES.TSInterfaceHeritage &&
                extendsExpression.expression.type === typescript_estree_1.AST_NODE_TYPES.Identifier) {
                // Extensions in the form of `interface A extends B`
                return extendsExpression.expression.name;
            }
            // Ignore interfaces that we don't understand
            this.logger.debug(`Ignored an interface expression of unknown type ${extendsExpression.expression.type} on ${declaration.id.name}`);
        })
            .filter(iface => Boolean(iface));
    }
    /**
     * Find the interface names of the given class.
     * @param declaration A class declaration.
     * @param fileName The file name of the current class.
     */
    getClassInterfaceNames(declaration, fileName) {
        const interfaceNames = [];
        if (declaration.implements) {
            for (const implement of declaration.implements) {
                if (implement.expression.type !== typescript_estree_1.AST_NODE_TYPES.Identifier) {
                    throw new Error(`Could not interpret the implements type on a class in ${fileName} on line ${implement.expression.loc.start.line} column ${implement.expression.loc.start.column}`);
                }
                interfaceNames.push(implement.expression.name);
            }
        }
        return interfaceNames;
    }
    /**
     * Load the referenced class, and obtain its full class declaration.
     * Classes can either be defined in this file (exported or not), or imported from another file.
     * @param classReference The reference to a class.
     * @param considerInterfaces If the class reference is allows to refer to an interface, as well as a class.
     */
    async loadClassDeclaration(classReference, considerInterfaces) {
        // Load the class as an AST
        let ast;
        try {
            ast = await this.resolutionContext.parseTypescriptFile(classReference.fileName);
        }
        catch (error) {
            throw new Error(`Could not load ${considerInterfaces ? 'class or interface' : 'class'} ${classReference.localName} from ${classReference.fileName}:\n${error.message}`);
        }
        const { exportedClasses, exportedInterfaces, declaredClasses, declaredInterfaces, importedElements, exportedImportedAll, exportedImportedElements, } = this.getClassElements(classReference.packageName, classReference.fileName, ast);
        // If the class has been exported in this file, return directly
        if (classReference.localName in exportedClasses) {
            const declaration = exportedClasses[classReference.localName];
            return this.enhanceLoadedWithComment(Object.assign(Object.assign({ type: 'class' }, classReference), { declaration,
                ast, abstract: declaration.abstract, generics: this.collectGenericTypes(declaration) }));
        }
        // If the class has been declared in this file, return directly
        if (classReference.localName in declaredClasses) {
            const declaration = declaredClasses[classReference.localName];
            return this.enhanceLoadedWithComment(Object.assign(Object.assign({ type: 'class' }, classReference), { declaration,
                ast, abstract: declaration.abstract, generics: this.collectGenericTypes(declaration) }));
        }
        // Only consider interfaces if explicitly enabled
        if (considerInterfaces) {
            // If the interface has been exported in this file, return directly
            if (classReference.localName in exportedInterfaces) {
                const declaration = exportedInterfaces[classReference.localName];
                return this.enhanceLoadedWithComment(Object.assign(Object.assign({ type: 'interface' }, classReference), { declaration,
                    ast, generics: this.collectGenericTypes(declaration) }));
            }
            // If the interface has been declared in this file, return directly
            if (classReference.localName in declaredInterfaces) {
                const declaration = declaredInterfaces[classReference.localName];
                return this.enhanceLoadedWithComment(Object.assign(Object.assign({ type: 'interface' }, classReference), { declaration,
                    ast, generics: this.collectGenericTypes(declaration) }));
            }
        }
        // If the class is available via an import, follow that import link
        if (classReference.localName in importedElements) {
            return this.loadClassDeclaration(importedElements[classReference.localName], considerInterfaces);
        }
        // If the class is available via an exported import, follow that import link
        if (classReference.localName in exportedImportedElements) {
            return this.loadClassDeclaration(exportedImportedElements[classReference.localName], considerInterfaces);
        }
        // If we still haven't found the class, iterate over all export all's
        for (const subFile of exportedImportedAll) {
            try {
                return await this.loadClassDeclaration(Object.assign({ localName: classReference.localName }, subFile), considerInterfaces);
            }
            catch (_a) {
                // Ignore class not found errors
            }
        }
        throw new Error(`Could not load ${considerInterfaces ? 'class or interface' : 'class'} ${classReference.localName} from ${classReference.fileName}`);
    }
    /**
     * Create a hash of generic types in the given class declaration.
     * @param classDeclaration A class or interface declaration.
     */
    collectGenericTypes(classDeclaration) {
        const genericTypes = {};
        if (classDeclaration.typeParameters) {
            for (const param of classDeclaration.typeParameters.params) {
                genericTypes[param.name.name] = { type: param.constraint };
            }
        }
        return genericTypes;
    }
    /**
     * Annotate the given loaded class or interface with a comment if it is present on the declaration.
     * @param classLoaded A loaded class or interface.
     */
    enhanceLoadedWithComment(classLoaded) {
        const commentData = new CommentLoader_1.CommentLoader({ classLoaded })
            .getCommentDataFromClassOrInterface(classLoaded.declaration);
        if (commentData.description) {
            classLoaded.comment = commentData.description;
        }
        return classLoaded;
    }
    /**
     * Load a class, and get all class elements from it.
     * @param packageName Package name we are importing from.
     * @param fileName A file path.
     */
    async loadClassElements(packageName, fileName) {
        const ast = await this.resolutionContext.parseTypescriptFile(fileName);
        return this.getClassElements(packageName, fileName, ast);
    }
    /**
     * Convert the given import path to an absolute file path, coupled with the module it is part of.
     * @param currentPackageName Package name we are importing from.
     * @param currentFilePath Absolute path to a file in which the import path occurs.
     * @param importPath Possibly relative path that is being imported.
     */
    importTargetToAbsolutePath(currentPackageName, currentFilePath, importPath) {
        // Handle import paths within the current package
        if (importPath.startsWith('.')) {
            return {
                packageName: currentPackageName,
                fileName: Path.join(Path.dirname(currentFilePath), importPath),
            };
        }
        // Handle import paths to other packages
        let packageName;
        let packagePath;
        if (importPath.startsWith('@')) {
            const slashIndexFirst = importPath.indexOf('/');
            if (slashIndexFirst < 0) {
                throw new Error(`Invalid scoped package name for import path '${importPath}' in '${currentFilePath}'`);
            }
            const slashIndexSecond = importPath.indexOf('/', slashIndexFirst + 1);
            if (slashIndexSecond < 0) {
                // Import form: "@scope/package"
                packageName = importPath;
            }
            else {
                // Import form: "@scope/package/path"
                packageName = importPath.slice(0, Math.max(0, slashIndexSecond));
                packagePath = importPath.slice(slashIndexSecond + 1);
            }
        }
        else {
            const slashIndex = importPath.indexOf('/');
            if (slashIndex < 0) {
                // Import form: "package"
                packageName = importPath;
            }
            else {
                // Import form: "package/path"
                packageName = importPath.slice(0, Math.max(0, slashIndex));
                packagePath = importPath.slice(slashIndex + 1);
            }
        }
        // Resolve paths
        const packageRoot = this.resolutionContext.resolvePackageIndex(packageName, currentFilePath);
        const remoteFilePath = packagePath ?
            Path.join(Path.dirname(packageRoot), packagePath) :
            packageRoot.slice(0, packageRoot.lastIndexOf('.'));
        return {
            packageName,
            fileName: remoteFilePath,
        };
    }
    /**
     * Get all class elements in a file.
     * @param packageName Package name we are importing from.
     * @param fileName A file path.
     * @param ast The parsed file.
     */
    getClassElements(packageName, fileName, ast) {
        const exportedClasses = {};
        const exportedInterfaces = {};
        const exportedImportedElements = {};
        const exportedImportedAll = [];
        const exportedUnknowns = {};
        const declaredClasses = {};
        const declaredInterfaces = {};
        const importedElements = {};
        for (const statement of ast.body) {
            if (statement.type === typescript_estree_1.AST_NODE_TYPES.ExportNamedDeclaration) {
                if (statement.declaration &&
                    statement.declaration.type === typescript_estree_1.AST_NODE_TYPES.ClassDeclaration) {
                    // Form: `export class A{}`
                    if (!statement.declaration.id) {
                        throw new Error(`Export parsing failure: missing exported class name in ${fileName} on line ${statement.declaration.loc.start.line} column ${statement.declaration.loc.start.column}`);
                    }
                    exportedClasses[statement.declaration.id.name] = statement.declaration;
                }
                else if (statement.declaration &&
                    statement.declaration.type === typescript_estree_1.AST_NODE_TYPES.TSInterfaceDeclaration) {
                    // Form: `export interface A{}`
                    exportedInterfaces[statement.declaration.id.name] = statement.declaration;
                }
                else if (statement.source &&
                    statement.source.type === typescript_estree_1.AST_NODE_TYPES.Literal &&
                    typeof statement.source.value === 'string') {
                    // Form: `export { A as B } from "b"`
                    for (const specifier of statement.specifiers) {
                        exportedImportedElements[specifier.exported.name] = Object.assign({ localName: specifier.local.name }, this.importTargetToAbsolutePath(packageName, fileName, statement.source.value));
                    }
                }
                else {
                    // Form: `export { A as B }`
                    for (const specifier of statement.specifiers) {
                        exportedUnknowns[specifier.exported.name] = specifier.local.name;
                    }
                }
            }
            else if (statement.type === typescript_estree_1.AST_NODE_TYPES.ExportAllDeclaration) {
                // Form: `export * from "b"`
                if (statement.source &&
                    statement.source.type === typescript_estree_1.AST_NODE_TYPES.Literal &&
                    typeof statement.source.value === 'string') {
                    exportedImportedAll.push(this.importTargetToAbsolutePath(packageName, fileName, statement.source.value));
                }
            }
            else if (statement.type === typescript_estree_1.AST_NODE_TYPES.ClassDeclaration && statement.id) {
                // Form: `declare class A {}`
                declaredClasses[statement.id.name] = statement;
            }
            else if (statement.type === typescript_estree_1.AST_NODE_TYPES.TSInterfaceDeclaration && statement.id) {
                // Form: `declare interface A {}`
                declaredInterfaces[statement.id.name] = statement;
            }
            else if (statement.type === typescript_estree_1.AST_NODE_TYPES.ImportDeclaration &&
                statement.source.type === typescript_estree_1.AST_NODE_TYPES.Literal &&
                typeof statement.source.value === 'string') {
                // Form: `import {A} from './lib/A'`
                for (const specifier of statement.specifiers) {
                    if (specifier.type === typescript_estree_1.AST_NODE_TYPES.ImportSpecifier) {
                        importedElements[specifier.local.name] = Object.assign({ localName: specifier.imported.name }, this.importTargetToAbsolutePath(packageName, fileName, statement.source.value));
                    }
                }
            }
        }
        return {
            exportedClasses,
            exportedInterfaces,
            exportedImportedElements,
            exportedImportedAll,
            exportedUnknowns,
            declaredClasses,
            declaredInterfaces,
            importedElements,
        };
    }
}
exports.ClassLoader = ClassLoader;
//# sourceMappingURL=ClassLoader.js.map