import type { ClassDeclaration, TSInterfaceDeclaration, MethodDefinition, TSPropertySignature, TSIndexSignature, BaseNode } from '@typescript-eslint/types/dist/ts-estree';
import type { ClassReference, ClassReferenceLoaded } from './ClassIndex';
import type { ParameterRangeUnresolved } from './ParameterLoader';
/**
 * Loads comments from fields in a given class.
 */
export declare class CommentLoader {
    private readonly classLoaded;
    constructor(args: CommentLoaderArgs);
    /**
     * Extract comment data from the given constructor.
     * @param constructor A constructor.
     */
    getCommentDataFromConstructor(constructor: MethodDefinition): ConstructorCommentData;
    /**
     * Get comment data from the given constructor comment.
     * @param comment A constructor comment string.
     * @param clazz A class reference, for error reporting.
     */
    static getCommentDataFromConstructorComment(comment: string, clazz: ClassReference): ConstructorCommentData;
    /**
     * Extract comment data from the given field.
     * @param field A field.
     */
    getCommentDataFromField(field: TSPropertySignature | TSIndexSignature): CommentData;
    /**
     * Extract comment data from the given class.
     * @param clazz A class or interface.
     */
    getCommentDataFromClassOrInterface(clazz: ClassDeclaration | TSInterfaceDeclaration): CommentData;
    /**
     * Get comment data from the given comment.
     * @param comment A comment string.
     * @param clazz A class reference, for error reporting.
     */
    static getCommentDataFromComment(comment: string, clazz: ClassReference): CommentData;
    /**
     * Get the comment string from the given node.
     * @param node A node, such as a field or constructor.
     */
    getCommentRaw(node: BaseNode): string | undefined;
}
export interface CommentLoaderArgs {
    classLoaded: ClassReferenceLoaded;
}
export declare type ConstructorCommentData = Record<string, CommentData>;
export interface CommentData {
    /**
     * The range of the parameter values.
     */
    range?: ParameterRangeUnresolved;
    /**
     * The default value.
     */
    default?: string;
    /**
     * If the field referenced by this comment should be ignored.
     */
    ignored?: boolean;
    /**
     * The human-readable description of this comment.
     */
    description?: string;
    /**
     * Parameters that were defined in this comment.
     */
    params?: Record<string, string>;
}
