"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CommentLoader = void 0;
const commentParse = require("comment-parser");
/**
 * Loads comments from fields in a given class.
 */
class CommentLoader {
    constructor(args) {
        this.classLoaded = args.classLoaded;
    }
    /**
     * Extract comment data from the given constructor.
     * @param constructor A constructor.
     */
    getCommentDataFromConstructor(constructor) {
        // Get the constructor comment
        const comment = this.getCommentRaw(constructor);
        if (comment) {
            return CommentLoader.getCommentDataFromConstructorComment(comment, this.classLoaded);
        }
        return {};
    }
    /**
     * Get comment data from the given constructor comment.
     * @param comment A constructor comment string.
     * @param clazz A class reference, for error reporting.
     */
    static getCommentDataFromConstructorComment(comment, clazz) {
        const data = {};
        // Iterate over all @param's
        const commentData = CommentLoader.getCommentDataFromComment(comment, clazz);
        if (commentData.params) {
            for (const [key, value] of Object.entries(commentData.params)) {
                data[key] = CommentLoader.getCommentDataFromComment(`/**${value.replace(/@/gu, '\n * @')}*/`, clazz);
            }
        }
        return data;
    }
    /**
     * Extract comment data from the given field.
     * @param field A field.
     */
    getCommentDataFromField(field) {
        const comment = this.getCommentRaw(field);
        if (comment) {
            return CommentLoader.getCommentDataFromComment(comment, this.classLoaded);
        }
        return {};
    }
    /**
     * Extract comment data from the given class.
     * @param clazz A class or interface.
     */
    getCommentDataFromClassOrInterface(clazz) {
        const comment = this.getCommentRaw(clazz);
        if (comment) {
            return CommentLoader.getCommentDataFromComment(comment, this.classLoaded);
        }
        return {};
    }
    /**
     * Get comment data from the given comment.
     * @param comment A comment string.
     * @param clazz A class reference, for error reporting.
     */
    static getCommentDataFromComment(comment, clazz) {
        const data = {};
        const commentParsed = commentParse(comment)[0];
        if (commentParsed) {
            // Extract description
            if (commentParsed.description.length > 0) {
                data.description = commentParsed.description.replace(/\n/gu, ' ');
            }
            // Extract tags
            for (const tag of commentParsed.tags) {
                switch (tag.tag.toLowerCase()) {
                    case 'range':
                        if (tag.type.length === 0) {
                            throw new Error(`Missing @range value {something} on a field in class ${clazz.localName} at ${clazz.fileName}`);
                        }
                        data.range = {
                            type: 'override',
                            value: tag.type,
                        };
                        break;
                    case 'default':
                        if (tag.type.length === 0) {
                            throw new Error(`Missing @default value {something} on a field in class ${clazz.localName} at ${clazz.fileName}`);
                        }
                        data.default = tag.type;
                        break;
                    case 'ignored':
                        data.ignored = true;
                        break;
                    case 'param':
                        if (!data.params) {
                            data.params = {};
                        }
                        data.params[tag.name] = tag.description;
                        if (data.params[tag.name].startsWith('- ')) {
                            data.params[tag.name] = data.params[tag.name].slice(2);
                        }
                        break;
                }
            }
        }
        return data;
    }
    /**
     * Get the comment string from the given node.
     * @param node A node, such as a field or constructor.
     */
    getCommentRaw(node) {
        const line = node.loc.start.line;
        for (const comment of this.classLoaded.ast.comments || []) {
            if (comment.loc.end.line === line - 1) {
                return `/*${comment.value}*/`;
            }
        }
    }
}
exports.CommentLoader = CommentLoader;
//# sourceMappingURL=CommentLoader.js.map