import type { Identifier, MethodDefinition, TSPropertySignature, TSTypeLiteral, TypeElement, TypeNode, TSIndexSignature, TSTypeReference, Parameter } from '@typescript-eslint/types/dist/ts-estree';
import type { ClassReferenceLoaded, InterfaceLoaded } from './ClassIndex';
import type { CommentData, ConstructorCommentData } from './CommentLoader';
import type { ConstructorData } from './ConstructorLoader';
/**
 * Interprets class parameters of a given class.
 */
export declare class ParameterLoader {
    private static readonly typeReferenceOverrides;
    private readonly classLoaded;
    private readonly commentLoader;
    constructor(args: ParameterLoaderArgs);
    /**
     * Load all parameter data from all fields in the given constructor.
     * @param constructor A constructor
     */
    loadConstructorFields(constructor: MethodDefinition): ConstructorData<ParameterRangeUnresolved>;
    /**
     * Load the parameter data from the given field in a constructor.
     * @param parameters The array of parameters that will be appended to.
     * @param constructorCommentData Comment data from the constructor.
     * @param field The field to load.
     */
    loadConstructorField(parameters: ParameterDataField<ParameterRangeUnresolved>[], constructorCommentData: ConstructorCommentData, field: Parameter): void;
    /**
     * Load all parameter data from all fields in the given interface.
     * If methods are found in the interface, an error is thrown.
     * @param iface An interface
     */
    loadInterfaceFields(iface: InterfaceLoaded): ParameterData<ParameterRangeUnresolved>[];
    /**
     * Load all parameter data from all fields in the given hash.
     * @param hash An hash element.
     */
    loadHashFields(hash: TSTypeLiteral): ParameterData<ParameterRangeUnresolved>[];
    /**
     * Load the parameter data from the given type element.
     * @param typeElement A type element, such as an interface or hash field.
     */
    loadTypeElementField(typeElement: TypeElement): ParameterData<ParameterRangeUnresolved> | undefined;
    /**
     * Load the parameter data from the given field.
     * @param field A field.
     * @param commentData Comment data about the given field.
     */
    loadField(field: Identifier | TSPropertySignature, commentData: CommentData): ParameterDataField<ParameterRangeUnresolved>;
    getFieldName(field: Identifier | TSPropertySignature): string;
    isFieldIndexedHash(field: Identifier | TSPropertySignature): boolean;
    isFieldUnique(field: Identifier | TSPropertySignature): boolean;
    isFieldRequired(field: Identifier | TSPropertySignature): boolean;
    getErrorIdentifierField(field: Identifier | TSPropertySignature): string;
    getErrorIdentifierIndex(): string;
    getRangeFromTypeNode(typeNode: TypeNode, errorIdentifier: string, nestedArrays?: number): ParameterRangeUnresolved;
    getFieldRange(field: Identifier | TSPropertySignature, commentData: CommentData): ParameterRangeUnresolved;
    getFieldDefault(commentData: CommentData): string | undefined;
    getFieldComment(commentData: CommentData): string | undefined;
    /**
     * Load the parameter data from the given index signature.
     * @param indexSignature An index signature.
     * @param commentData Comment data about the given field.
     */
    loadIndex(indexSignature: TSIndexSignature, commentData: CommentData): ParameterDataIndex<ParameterRangeUnresolved>;
    getIndexDomain(indexSignature: TSIndexSignature): 'string' | 'number' | 'boolean';
    getIndexRange(indexSignature: TSIndexSignature, commentData: CommentData): ParameterRangeUnresolved;
    /**
     * Iterate over all type reference override handler to see if one of them overrides the given type.
     * @param typeNode A type reference node.
     */
    handleTypeOverride(typeNode: TSTypeReference): ParameterRangeUnresolved | undefined;
}
export interface ParameterLoaderArgs {
    classLoaded: ClassReferenceLoaded;
}
export declare type ParameterData<R> = ParameterDataField<R> | ParameterDataIndex<R>;
export interface ParameterDataField<R> {
    /**
     * The data type.
     */
    type: 'field';
    /**
     * The parameter name.
     */
    name: string;
    /**
     * If only one value for the given parameter can exist.
     * This is always false for an array.
     */
    unique: boolean;
    /**
     * If the parameter MUST have a value.
     */
    required: boolean;
    /**
     * The range of the parameter values.
     */
    range: R;
    /**
     * The default value.
     */
    default?: string;
    /**
     * The human-readable description of this parameter.
     */
    comment?: string;
}
export interface ParameterDataIndex<R> {
    /**
     * The data type.
     */
    type: 'index';
    /**
     * The domain of the parameter keys.
     */
    domain: 'string' | 'number' | 'boolean';
    /**
     * The range of the parameter values.
     */
    range: R;
    /**
     * The default value.
     */
    default?: string;
    /**
     * The human-readable description of this parameter.
     */
    comment?: string;
}
export declare type ParameterRangeUnresolved = {
    type: 'raw';
    value: 'boolean' | 'number' | 'string';
} | {
    type: 'override';
    value: string;
} | {
    type: 'interface';
    value: string;
} | {
    type: 'hash';
    value: TSTypeLiteral;
} | {
    type: 'undefined';
};
export declare type ParameterRangeResolved = {
    type: 'raw';
    value: 'boolean' | 'number' | 'string';
} | {
    type: 'override';
    value: string;
} | {
    type: 'class';
    value: ClassReferenceLoaded;
} | {
    type: 'nested';
    value: ParameterData<ParameterRangeResolved>[];
} | {
    type: 'undefined';
};
