"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ParameterLoader = void 0;
const typescript_estree_1 = require("@typescript-eslint/typescript-estree");
const CommentLoader_1 = require("./CommentLoader");
const TypeReferenceOverrideAliasRecord_1 = require("./typereferenceoverride/TypeReferenceOverrideAliasRecord");
/**
 * Interprets class parameters of a given class.
 */
class ParameterLoader {
    constructor(args) {
        this.classLoaded = args.classLoaded;
        this.commentLoader = new CommentLoader_1.CommentLoader({ classLoaded: this.classLoaded });
    }
    /**
     * Load all parameter data from all fields in the given constructor.
     * @param constructor A constructor
     */
    loadConstructorFields(constructor) {
        // Load the constructor comment
        const constructorCommentData = this.commentLoader.getCommentDataFromConstructor(constructor);
        // Load all constructor parameters
        const parameters = [];
        for (const field of constructor.value.params) {
            this.loadConstructorField(parameters, constructorCommentData, field);
        }
        return { parameters };
    }
    /**
     * Load the parameter data from the given field in a constructor.
     * @param parameters The array of parameters that will be appended to.
     * @param constructorCommentData Comment data from the constructor.
     * @param field The field to load.
     */
    loadConstructorField(parameters, constructorCommentData, field) {
        if (field.type === typescript_estree_1.AST_NODE_TYPES.Identifier) {
            const commentData = constructorCommentData[field.name] || {};
            if (!commentData.ignored) {
                parameters.push(this.loadField(field, commentData));
            }
        }
        else if (field.type === typescript_estree_1.AST_NODE_TYPES.TSParameterProperty) {
            this.loadConstructorField(parameters, constructorCommentData, field.parameter);
        }
        else {
            throw new Error(`Could not understand constructor parameter type ${field.type} in ${this.classLoaded.localName} at ${this.classLoaded.fileName}`);
        }
    }
    /**
     * Load all parameter data from all fields in the given interface.
     * If methods are found in the interface, an error is thrown.
     * @param iface An interface
     */
    loadInterfaceFields(iface) {
        return iface.declaration.body.body
            .map(field => this.loadTypeElementField(field))
            .filter(Boolean);
    }
    /**
     * Load all parameter data from all fields in the given hash.
     * @param hash An hash element.
     */
    loadHashFields(hash) {
        return hash.members
            .map(field => this.loadTypeElementField(field))
            .filter(Boolean);
    }
    /**
     * Load the parameter data from the given type element.
     * @param typeElement A type element, such as an interface or hash field.
     */
    loadTypeElementField(typeElement) {
        let commentData;
        switch (typeElement.type) {
            case typescript_estree_1.AST_NODE_TYPES.TSPropertySignature:
                commentData = this.commentLoader.getCommentDataFromField(typeElement);
                if (!commentData.ignored) {
                    return this.loadField(typeElement, commentData);
                }
                return;
            case typescript_estree_1.AST_NODE_TYPES.TSIndexSignature:
                commentData = this.commentLoader.getCommentDataFromField(typeElement);
                if (!commentData.ignored) {
                    return this.loadIndex(typeElement, commentData);
                }
                return;
            default:
                throw new Error(`Unsupported field type ${typeElement.type} in ${this.classLoaded.localName} in ${this.classLoaded.fileName}`);
        }
    }
    /**
     * Load the parameter data from the given field.
     * @param field A field.
     * @param commentData Comment data about the given field.
     */
    loadField(field, commentData) {
        // Required data
        const parameterData = {
            type: 'field',
            name: this.getFieldName(field),
            unique: this.isFieldUnique(field),
            required: this.isFieldRequired(field),
            range: this.getFieldRange(field, commentData),
        };
        // Optional data
        const defaultValue = this.getFieldDefault(commentData);
        if (defaultValue) {
            parameterData.default = defaultValue;
        }
        const comment = this.getFieldComment(commentData);
        if (comment) {
            parameterData.comment = comment;
        }
        return parameterData;
    }
    getFieldName(field) {
        if ('name' in field) {
            // If Identifier
            return field.name;
        }
        // Else TSPropertySignature
        if (field.key.type === typescript_estree_1.AST_NODE_TYPES.Identifier) {
            return field.key.name;
        }
        throw new Error(`Unsupported field key type ${field.key.type} in interface ${this.classLoaded.localName} in ${this.classLoaded.fileName}`);
    }
    isFieldIndexedHash(field) {
        return Boolean(field.typeAnnotation &&
            field.typeAnnotation.typeAnnotation.type === typescript_estree_1.AST_NODE_TYPES.TSTypeLiteral &&
            field.typeAnnotation.typeAnnotation.members.some(member => member.type === typescript_estree_1.AST_NODE_TYPES.TSIndexSignature));
    }
    isFieldUnique(field) {
        return !(field.typeAnnotation && field.typeAnnotation.typeAnnotation.type === typescript_estree_1.AST_NODE_TYPES.TSArrayType) &&
            !this.isFieldIndexedHash(field);
    }
    isFieldRequired(field) {
        return !field.optional && !this.isFieldIndexedHash(field);
    }
    getErrorIdentifierField(field) {
        return `field ${this.getFieldName(field)}`;
    }
    getErrorIdentifierIndex() {
        return `an index signature`;
    }
    getRangeFromTypeNode(typeNode, errorIdentifier, nestedArrays = 0) {
        // Don't allow arrays to be nested
        if (nestedArrays > 1) {
            throw new Error(`Detected illegal nested array type for ${errorIdentifier} in ${this.classLoaded.localName} at ${this.classLoaded.fileName}`);
        }
        let typeAliasOverride;
        // eslint-disable-next-line @typescript-eslint/switch-exhaustiveness-check
        switch (typeNode.type) {
            case typescript_estree_1.AST_NODE_TYPES.TSTypeReference:
                if (typeNode.typeName.type === typescript_estree_1.AST_NODE_TYPES.Identifier) {
                    // TS raw types can also start with an uppercase
                    switch (typeNode.typeName.name) {
                        case 'Boolean':
                            return { type: 'raw', value: 'boolean' };
                        case 'Number':
                            return { type: 'raw', value: 'number' };
                        case 'String':
                            return { type: 'raw', value: 'string' };
                        case 'Array':
                            if (typeNode.typeParameters && typeNode.typeParameters.params.length === 1) {
                                return this.getRangeFromTypeNode(typeNode.typeParameters.params[0], errorIdentifier, nestedArrays + 1);
                            }
                            throw new Error(`Found invalid Array field type at ${errorIdentifier} in ${this.classLoaded.localName} at ${this.classLoaded.fileName}`);
                        default:
                            // First check if the type is be a generic type
                            if (typeNode.typeName.name in this.classLoaded.generics) {
                                const genericProperties = this.classLoaded.generics[typeNode.typeName.name];
                                if (!genericProperties.type) {
                                    throw new Error(`Found untyped generic field type at ${errorIdentifier} in ${this.classLoaded.localName} at ${this.classLoaded.fileName}`);
                                }
                                return this.getRangeFromTypeNode(genericProperties.type, errorIdentifier);
                            }
                            // Check if this node is a predefined type alias
                            typeAliasOverride = this.handleTypeOverride(typeNode);
                            if (typeAliasOverride) {
                                return typeAliasOverride;
                            }
                            // Otherwise, assume we have an interface/class parameter
                            return { type: 'interface', value: typeNode.typeName.name };
                    }
                }
                break;
            case typescript_estree_1.AST_NODE_TYPES.TSArrayType:
                return this.getRangeFromTypeNode(typeNode.elementType, errorIdentifier, nestedArrays + 1);
            case typescript_estree_1.AST_NODE_TYPES.TSBooleanKeyword:
                return { type: 'raw', value: 'boolean' };
            case typescript_estree_1.AST_NODE_TYPES.TSNumberKeyword:
                return { type: 'raw', value: 'number' };
            case typescript_estree_1.AST_NODE_TYPES.TSStringKeyword:
                return { type: 'raw', value: 'string' };
            case typescript_estree_1.AST_NODE_TYPES.TSTypeLiteral:
                return { type: 'hash', value: typeNode };
            case typescript_estree_1.AST_NODE_TYPES.TSUnknownKeyword:
            case typescript_estree_1.AST_NODE_TYPES.TSUndefinedKeyword:
            case typescript_estree_1.AST_NODE_TYPES.TSVoidKeyword:
            case typescript_estree_1.AST_NODE_TYPES.TSNullKeyword:
            case typescript_estree_1.AST_NODE_TYPES.TSAnyKeyword:
            case typescript_estree_1.AST_NODE_TYPES.TSUnionType:
            case typescript_estree_1.AST_NODE_TYPES.TSTupleType:
                return { type: 'undefined' };
        }
        throw new Error(`Could not understand parameter type ${typeNode.type} of ${errorIdentifier} in ${this.classLoaded.localName} at ${this.classLoaded.fileName}`);
    }
    getFieldRange(field, commentData) {
        // Check comment data
        if (commentData.range) {
            return commentData.range;
        }
        // Check the typescript raw field type
        if (field.typeAnnotation) {
            return this.getRangeFromTypeNode(field.typeAnnotation.typeAnnotation, this.getErrorIdentifierField(field));
        }
        throw new Error(`Missing field type on ${this.getFieldName(field)} in ${this.classLoaded.localName} at ${this.classLoaded.fileName}`);
    }
    getFieldDefault(commentData) {
        return commentData.default;
    }
    getFieldComment(commentData) {
        return commentData.description;
    }
    /**
     * Load the parameter data from the given index signature.
     * @param indexSignature An index signature.
     * @param commentData Comment data about the given field.
     */
    loadIndex(indexSignature, commentData) {
        // Required data
        const parameterData = {
            type: 'index',
            domain: this.getIndexDomain(indexSignature),
            range: this.getIndexRange(indexSignature, commentData),
        };
        // Optional data
        const defaultValue = this.getFieldDefault(commentData);
        if (defaultValue) {
            parameterData.default = defaultValue;
        }
        const comment = this.getFieldComment(commentData);
        if (comment) {
            parameterData.comment = comment;
        }
        return parameterData;
    }
    getIndexDomain(indexSignature) {
        if (indexSignature.parameters.length !== 1) {
            throw new Error(`Expected exactly one key in index signature in ${this.classLoaded.localName} at ${this.classLoaded.fileName}`);
        }
        if (indexSignature.parameters[0].type !== 'Identifier') {
            throw new Error(`Only identifier-based index signatures are allowed in ${this.classLoaded.localName} at ${this.classLoaded.fileName}`);
        }
        if (!indexSignature.parameters[0].typeAnnotation) {
            throw new Error(`Missing key type annotation in index signature in ${this.classLoaded.localName} at ${this.classLoaded.fileName}`);
        }
        const type = this.getRangeFromTypeNode(indexSignature.parameters[0].typeAnnotation.typeAnnotation, this.getErrorIdentifierIndex());
        if (type.type !== 'raw') {
            throw new Error(`Only raw types are allowed in index signature keys in ${this.classLoaded.localName} at ${this.classLoaded.fileName}`);
        }
        return type.value;
    }
    getIndexRange(indexSignature, commentData) {
        // Check comment data
        if (commentData.range) {
            return commentData.range;
        }
        // Check the typescript raw field type
        if (indexSignature.typeAnnotation) {
            return this.getRangeFromTypeNode(indexSignature.typeAnnotation.typeAnnotation, this.getErrorIdentifierIndex());
        }
        throw new Error(`Missing field type on ${this.getErrorIdentifierIndex()} in ${this.classLoaded.localName} at ${this.classLoaded.fileName}`);
    }
    /**
     * Iterate over all type reference override handler to see if one of them overrides the given type.
     * @param typeNode A type reference node.
     */
    handleTypeOverride(typeNode) {
        for (const typeReferenceOverride of ParameterLoader.typeReferenceOverrides) {
            const handled = typeReferenceOverride.handle(typeNode);
            if (handled) {
                return handled;
            }
        }
    }
}
exports.ParameterLoader = ParameterLoader;
ParameterLoader.typeReferenceOverrides = [
    new TypeReferenceOverrideAliasRecord_1.TypeReferenceOverrideAliasRecord(),
];
//# sourceMappingURL=ParameterLoader.js.map