import type { TSTypeLiteral } from '@typescript-eslint/types/dist/ts-estree';
import type { ClassIndex, ClassLoaded, ClassReference, ClassReferenceLoaded, InterfaceLoaded } from './ClassIndex';
import type { ClassLoader } from './ClassLoader';
import type { ConstructorData } from './ConstructorLoader';
import type { ParameterData, ParameterDataField, ParameterRangeResolved, ParameterRangeUnresolved } from './ParameterLoader';
export declare class ParameterResolver {
    private readonly classLoader;
    private readonly ignoreClasses;
    constructor(args: ParameterResolverArgs);
    /**
     * Resolve all constructor parameters of a given constructor index.
     * @param unresolvedParametersIndex An index of unresolved constructor data.
     * @param classIndex A class index.
     */
    resolveAllConstructorParameters(unresolvedParametersIndex: ClassIndex<ConstructorData<ParameterRangeUnresolved>>, classIndex: ClassIndex<ClassReferenceLoaded>): Promise<ClassIndex<ConstructorData<ParameterRangeResolved>>>;
    /**
     * Resolve all parameters of a given constructor.
     * @param unresolvedConstructorData Unresolved constructor data.
     * @param owningClass The class in which the given constructor is declared.
     */
    resolveConstructorParameters(unresolvedConstructorData: ConstructorData<ParameterRangeUnresolved>, owningClass: ClassLoaded): Promise<ConstructorData<ParameterRangeResolved>>;
    /**
     * Resolve the given array of parameter data in parallel.
     * @param parameters An array of unresolved parameters.
     * @param owningClass The class in which the given parameters are declared.
     */
    resolveParameterData(parameters: ParameterData<ParameterRangeUnresolved>[], owningClass: ClassReferenceLoaded): Promise<ParameterDataField<ParameterRangeResolved>[]>;
    /**
     * Resolve the given parameter range.
     * @param range An unresolved parameter range.
     * @param owningClass The class this range was defined in.
     */
    resolveRange(range: ParameterRangeUnresolved, owningClass: ClassReferenceLoaded): Promise<ParameterRangeResolved>;
    /**
     * Resolve a class or interface.
     * @param interfaceName A class or interface name.
     * @param owningClass The class this interface was used in.
     */
    resolveRangeInterface(interfaceName: string, owningClass: ClassReferenceLoaded): Promise<ParameterRangeResolved>;
    /**
     * Check if the given interface should actually be considered a class.
     * Concretely, it will check whether or not there is at least one method or constructor.
     * @param iface An interface.
     */
    isInterfaceImplicitClass(iface: InterfaceLoaded): boolean;
    /**
     * Load the given class reference, which could either be a class or interface.
     *
     * If it's a class, just load it, without loading its superclasses.
     * If it's an interface, load all of its superinterfaces recursively.
     *
     * This method will throw if an interface extends from a class.
     *
     * @param classReference A class reference.
     */
    loadClassOrInterfacesChain(classReference: ClassReference): Promise<ClassLoaded | InterfaceLoaded>;
    /**
     * Recursively get all fields from the given interface.
     * @param iface A loaded interface.
     */
    getNestedFieldsFromInterface(iface: InterfaceLoaded): Promise<ParameterData<ParameterRangeResolved>[]>;
    /**
     * Recursively get all fields from the given hash.
     * @param hash A hash object.
     * @param owningClass The class this hash is declared in.
     */
    getNestedFieldsFromHash(hash: TSTypeLiteral, owningClass: ClassReferenceLoaded): Promise<ParameterData<ParameterRangeResolved>[]>;
}
export interface ParameterResolverArgs {
    classLoader: ClassLoader;
    ignoreClasses: Record<string, boolean>;
}
