"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ParameterResolver = void 0;
const typescript_estree_1 = require("@typescript-eslint/typescript-estree");
const ParameterLoader_1 = require("./ParameterLoader");
class ParameterResolver {
    constructor(args) {
        this.classLoader = args.classLoader;
        this.ignoreClasses = args.ignoreClasses;
    }
    /**
     * Resolve all constructor parameters of a given constructor index.
     * @param unresolvedParametersIndex An index of unresolved constructor data.
     * @param classIndex A class index.
     */
    async resolveAllConstructorParameters(unresolvedParametersIndex, classIndex) {
        const resolvedParametersIndex = {};
        // Resolve parameters for the different constructors in parallel
        await Promise.all(Object.entries(unresolvedParametersIndex)
            .map(async ([className, parameters]) => {
            const classOrInterface = classIndex[className];
            if (classOrInterface.type === 'class') {
                resolvedParametersIndex[className] = await this.resolveConstructorParameters(parameters, classOrInterface);
            }
        }));
        return resolvedParametersIndex;
    }
    /**
     * Resolve all parameters of a given constructor.
     * @param unresolvedConstructorData Unresolved constructor data.
     * @param owningClass The class in which the given constructor is declared.
     */
    async resolveConstructorParameters(unresolvedConstructorData, owningClass) {
        return {
            parameters: await this.resolveParameterData(unresolvedConstructorData.parameters, owningClass),
        };
    }
    /**
     * Resolve the given array of parameter data in parallel.
     * @param parameters An array of unresolved parameters.
     * @param owningClass The class in which the given parameters are declared.
     */
    async resolveParameterData(parameters, owningClass) {
        return await Promise.all(parameters
            .map(async (parameter) => (Object.assign(Object.assign({}, parameter), { range: await this.resolveRange(parameter.range, owningClass) }))));
    }
    /**
     * Resolve the given parameter range.
     * @param range An unresolved parameter range.
     * @param owningClass The class this range was defined in.
     */
    async resolveRange(range, owningClass) {
        switch (range.type) {
            case 'raw':
            case 'override':
                return range;
            case 'interface':
                if (range.value in this.ignoreClasses) {
                    return {
                        type: 'undefined',
                    };
                }
                return await this.resolveRangeInterface(range.value, owningClass);
            case 'hash':
                return {
                    type: 'nested',
                    value: await this.getNestedFieldsFromHash(range.value, owningClass),
                };
            case 'undefined':
                return {
                    type: 'undefined',
                };
        }
    }
    /**
     * Resolve a class or interface.
     * @param interfaceName A class or interface name.
     * @param owningClass The class this interface was used in.
     */
    async resolveRangeInterface(interfaceName, owningClass) {
        const classOrInterface = await this.loadClassOrInterfacesChain({
            packageName: owningClass.packageName,
            localName: interfaceName,
            fileName: owningClass.fileName,
        });
        // If we find a class, or an interface that is implicitly a class, return the class reference directly
        if (classOrInterface.type === 'class' ||
            (classOrInterface.type === 'interface' && this.isInterfaceImplicitClass(classOrInterface))) {
            return {
                type: 'class',
                value: classOrInterface,
            };
        }
        // If we find an interface, load it as a hash with nested fields
        return {
            type: 'nested',
            value: await this.getNestedFieldsFromInterface(classOrInterface),
        };
    }
    /**
     * Check if the given interface should actually be considered a class.
     * Concretely, it will check whether or not there is at least one method or constructor.
     * @param iface An interface.
     */
    isInterfaceImplicitClass(iface) {
        for (const field of iface.declaration.body.body) {
            if (field.type === typescript_estree_1.AST_NODE_TYPES.TSMethodSignature ||
                field.type === typescript_estree_1.AST_NODE_TYPES.TSConstructSignatureDeclaration ||
                (field.type === typescript_estree_1.AST_NODE_TYPES.TSPropertySignature && field.typeAnnotation &&
                    field.typeAnnotation.typeAnnotation.type === typescript_estree_1.AST_NODE_TYPES.TSFunctionType)) {
                return true;
            }
        }
        return false;
    }
    /**
     * Load the given class reference, which could either be a class or interface.
     *
     * If it's a class, just load it, without loading its superclasses.
     * If it's an interface, load all of its superinterfaces recursively.
     *
     * This method will throw if an interface extends from a class.
     *
     * @param classReference A class reference.
     */
    async loadClassOrInterfacesChain(classReference) {
        const classOrInterface = await this.classLoader.loadClassDeclaration(classReference, true);
        // If the result is an interface, load all its super interfaces recursively
        if (classOrInterface.type === 'interface') {
            classOrInterface.superInterfaces = await Promise.all(this.classLoader
                .getSuperInterfaceNames(classOrInterface.declaration, classOrInterface.fileName)
                .filter(interfaceName => !(interfaceName in this.ignoreClasses))
                .map(async (interfaceName) => {
                const superInterface = await this.loadClassOrInterfacesChain({
                    packageName: classOrInterface.packageName,
                    localName: interfaceName,
                    fileName: classOrInterface.fileName,
                });
                if (superInterface.type !== 'interface') {
                    throw new Error(`Detected interface ${classOrInterface.localName} extending from a class ${interfaceName} in ${classReference.fileName}`);
                }
                return superInterface;
            }));
        }
        return classOrInterface;
    }
    /**
     * Recursively get all fields from the given interface.
     * @param iface A loaded interface.
     */
    async getNestedFieldsFromInterface(iface) {
        const parameterLoader = new ParameterLoader_1.ParameterLoader({ classLoaded: iface });
        const unresolvedFields = parameterLoader.loadInterfaceFields(iface);
        return this.resolveParameterData(unresolvedFields, iface);
    }
    /**
     * Recursively get all fields from the given hash.
     * @param hash A hash object.
     * @param owningClass The class this hash is declared in.
     */
    async getNestedFieldsFromHash(hash, owningClass) {
        const parameterLoader = new ParameterLoader_1.ParameterLoader({ classLoaded: owningClass });
        const unresolvedFields = parameterLoader.loadHashFields(hash);
        return this.resolveParameterData(unresolvedFields, owningClass);
    }
}
exports.ParameterResolver = ParameterResolver;
//# sourceMappingURL=ParameterResolver.js.map