"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ExternalModulesLoader = void 0;
const componentsjs_1 = require("componentsjs");
/**
 * Loads components from Node modules.
 */
class ExternalModulesLoader {
    constructor(args) {
        this.pathDestination = args.pathDestination;
        this.packageMetadata = args.packageMetadata;
        this.logger = args.logger;
    }
    /**
     * Determine packages referred to by components.
     * @param classIndex The available classes.
     * @param constructors The available constructors.
     */
    findExternalPackages(classIndex, constructors) {
        const externalPackages = {};
        // Handle super classes
        for (const classReference of Object.values(classIndex)) {
            this.indexClassInExternalPackage(classReference, externalPackages);
        }
        // Handle constructor parameters
        for (const ctor of Object.values(constructors)) {
            for (const parameter of ctor.parameters) {
                this.indexParameterInExternalPackage(parameter, externalPackages);
            }
        }
        return Object.keys(externalPackages);
    }
    indexClassInExternalPackage(classReference, externalPackages) {
        if (classReference.packageName !== this.packageMetadata.name) {
            externalPackages[classReference.packageName] = true;
        }
        if (classReference.type === 'class') {
            if (classReference.superClass) {
                this.indexClassInExternalPackage(classReference.superClass, externalPackages);
            }
            if (classReference.implementsInterfaces) {
                for (const iface of classReference.implementsInterfaces) {
                    this.indexClassInExternalPackage(iface, externalPackages);
                }
            }
        }
    }
    indexParameterInExternalPackage(parameter, externalPackages) {
        switch (parameter.range.type) {
            case 'raw':
            case 'override':
            case 'undefined':
                break;
            case 'class':
                this.indexClassInExternalPackage(parameter.range.value, externalPackages);
                break;
            case 'nested':
                for (const nestedParameter of parameter.range.value) {
                    this.indexParameterInExternalPackage(nestedParameter, externalPackages);
                }
                break;
        }
    }
    /**
     * Build the module state.
     * Adapted from {@link ModuleStateBuilder#buildModuleState}.
     *
     * Instead of loading the modules for all available packages,
     * only the packages with their name included in the packageNames array are included.
     * This leads to better performance, as we usually need only a small subset of all packages.
     *
     * @param req The `require` instance.
     * @param packageNames Names of the packages to load.
     */
    async buildModuleStateSelective(req, packageNames) {
        const moduleStateBuilder = new componentsjs_1.ModuleStateBuilder(this.logger);
        const mainModulePath = this.pathDestination.packageRootDirectory;
        const nodeModuleImportPaths = moduleStateBuilder.buildNodeModuleImportPaths(mainModulePath);
        const nodeModulePaths = this.buildNodeModulePathsSelective(req, nodeModuleImportPaths, packageNames);
        const packageJsons = await moduleStateBuilder.buildPackageJsons(nodeModulePaths);
        await moduleStateBuilder.preprocessPackageJsons(packageJsons);
        const componentModules = await moduleStateBuilder.buildComponentModules(packageJsons);
        const contexts = await moduleStateBuilder.buildComponentContexts(packageJsons);
        const importPaths = await moduleStateBuilder.buildComponentImportPaths(packageJsons);
        return {
            mainModulePath,
            nodeModuleImportPaths,
            nodeModulePaths,
            packageJsons,
            componentModules,
            contexts,
            importPaths,
        };
    }
    /**
     * Get all currently available node module paths.
     * Adapted from {@link ModuleStateBuilder#buildNodeModulePaths}.
     * @param req The `require` instance.
     * @param nodeModuleImportPaths The import paths to resolve from.
     * @param packageNames The package names to resolve the path for.
     */
    buildNodeModulePathsSelective(req, nodeModuleImportPaths, packageNames) {
        return packageNames
            .map(packageName => req
            .resolve(`${packageName}/package.json`, { paths: nodeModuleImportPaths })
            .slice(0, -13));
    }
    /**
     * Loads all components from the given Node package names.
     * @param req The `require` instance.
     * @param externalPackages
     */
    async loadExternalComponents(req, externalPackages) {
        // Load module state for the external packages
        const moduleState = await this.buildModuleStateSelective(req, externalPackages);
        // Load components for this module state (code inspired by ComponentsManagerBuilder from Components.js)
        const componentResources = {};
        const objectLoader = componentsjs_1.ComponentsManagerBuilder.createObjectLoader();
        const componentRegistry = new componentsjs_1.ComponentRegistry({
            moduleState,
            objectLoader,
            logger: this.logger,
            componentResources,
            skipContextValidation: true,
        });
        await componentRegistry.registerAvailableModules();
        const componentFinalizer = new componentsjs_1.ComponentRegistryFinalizer({
            objectLoader,
            logger: this.logger,
            componentResources,
            componentRegistry,
        });
        componentFinalizer.finalize();
        // Index available package.json by package name
        const packageJsons = {};
        for (const [path, packageJson] of Object.entries(moduleState.packageJsons)) {
            packageJsons[packageJson.name] = { contents: packageJson, path };
        }
        // Index all loaded components
        const externalComponents = {
            moduleState,
            components: {},
        };
        for (const componentResource of Object.values(componentResources)) {
            const packageName = componentResource.property.module.property.requireName.value;
            // Initialize metadata for a package if it doesn't exist yet
            if (!externalComponents.components[packageName]) {
                const packageJson = packageJsons[packageName];
                if (!packageJson) {
                    this.logger.warn(`Could not find a package.json for '${packageName}'`);
                }
                else {
                    const contexts = packageJson.contents['lsd:contexts'];
                    externalComponents.components[packageName] = {
                        contextIris: Object.keys(contexts),
                        componentNamesToIris: {},
                    };
                }
            }
            // Add component to package
            if (externalComponents.components[packageName]) {
                externalComponents.components[packageName]
                    .componentNamesToIris[componentResource.property.requireElement.value] = componentResource.value;
            }
        }
        return externalComponents;
    }
}
exports.ExternalModulesLoader = ExternalModulesLoader;
//# sourceMappingURL=ExternalModulesLoader.js.map