import type { ContextParser, JsonLdContextNormalized } from 'jsonld-context-parser';
import type { ClassIndex, ClassLoaded, ClassReference, ClassReferenceLoaded } from '../parse/ClassIndex';
import type { ConstructorData } from '../parse/ConstructorLoader';
import type { PackageMetadata } from '../parse/PackageMetadataLoader';
import type { ParameterData, ParameterRangeResolved } from '../parse/ParameterLoader';
import type { ExternalComponents } from '../resolution/ExternalModulesLoader';
import type { ComponentDefinition, ComponentDefinitions, ComponentDefinitionsIndex, ConstructorArgumentDefinition, ConstructorFieldDefinition, ParameterDefinition } from './ComponentDefinitions';
import type { ContextConstructor } from './ContextConstructor';
/**
 * Creates declarative JSON components for the given classes.
 */
export declare class ComponentConstructor {
    private readonly packageMetadata;
    private readonly contextConstructor;
    private readonly pathDestination;
    private readonly classAndInterfaceIndex;
    private readonly classConstructors;
    private readonly externalComponents;
    private readonly contextParser;
    constructor(args: ComponentConstructorArgs);
    /**
     * Construct component definitions for all classes in the current index.
     */
    constructComponents(): Promise<ComponentDefinitions>;
    /**
     * Construct a component definitions index.
     * @param definitions The component definitions for which the index should be constructed.
     * @param fileExtension The file extension to apply on files.
     */
    constructComponentsIndex(definitions: ComponentDefinitions, fileExtension: string): Promise<ComponentDefinitionsIndex>;
    /**
     * Determine the relative path of a component file within a package.
     * @param sourcePath The absolute path to a class file.
     */
    getPathRelative(sourcePath: string): string;
    /**
     * Determine the path a component file should exist at based on a class source file path.
     * @param sourcePath The absolute path to a class file.
     */
    getPathDestination(sourcePath: string): string;
    /**
     * Determine the IRI of the given source path.
     * @param sourcePath The relative path to a components file.
     */
    getImportPathIri(sourcePath: string): string;
    /**
     * Construct a component definition from the given constructor data.
     * @param context A parsed JSON-LD context.
     * @param externalContextsCallback Callback for external contexts.
     * @param classReference Class reference of the class component.
     * @param constructorData Constructor data of the owning class.
     */
    constructComponent(context: JsonLdContextNormalized, externalContextsCallback: ExternalContextCallback, classReference: ClassReferenceLoaded, constructorData: ConstructorData<ParameterRangeResolved>): Promise<ComponentDefinition>;
    /**
     * Construct a compacted module IRI.
     * @param context A parsed JSON-LD context.
     */
    moduleIriToId(context: JsonLdContextNormalized): string;
    /**
     * Construct a compacted class IRI.
     * @param context A parsed JSON-LD context.
     * @param externalContextsCallback Callback for external contexts.
     * @param classReference The class reference.
     */
    classNameToId(context: JsonLdContextNormalized, externalContextsCallback: ExternalContextCallback, classReference: ClassReference): Promise<string>;
    /**
     * Construct a compacted field IRI.
     * @param context A parsed JSON-LD context.
     * @param classReference The class reference.
     * @param fieldName The name of the field.
     * @param scope The current field scope.
     */
    fieldNameToId(context: JsonLdContextNormalized, classReference: ClassReference, fieldName: string, scope: FieldScope): string;
    /**
     * Construct constructor arguments from the given constructor data.
     * Additionally, parameters will be appended to the parameters array.
     *
     * @param context A parsed JSON-LD context.
     * @param externalContextsCallback Callback for external contexts.
     * @param classReference Class reference of the class component owning this constructor.
     * @param constructorData Constructor data of the owning class.
     * @param parameters The array of parameters of the owning class, which will be appended to.
     */
    constructParameters(context: JsonLdContextNormalized, externalContextsCallback: ExternalContextCallback, classReference: ClassLoaded, constructorData: ConstructorData<ParameterRangeResolved>, parameters: ParameterDefinition[]): Promise<ConstructorArgumentDefinition[]>;
    /**
     * Construct a constructor argument from the given parameter data.
     * Additionally, one (or more) parameters will be appended to the parameters array.
     *
     * This may be invoked recursively based on the parameter type.
     *
     * @param context A parsed JSON-LD context.
     * @param externalContextsCallback Callback for external contexts.
     * @param classReference Class reference of the class component owning this parameter.
     * @param parameterData Parameter data.
     * @param parameters The array of parameters of the owning class, which will be appended to.
     * @param fieldId The @id of the field.
     * @param scope The current field scope.
     */
    parameterDataToConstructorArgument(context: JsonLdContextNormalized, externalContextsCallback: ExternalContextCallback, classReference: ClassLoaded, parameterData: ParameterData<ParameterRangeResolved>, parameters: ParameterDefinition[], fieldId: string, scope: FieldScope): Promise<ConstructorArgumentDefinition>;
    /**
     * For the given parameter with nested range, construct field definitions for all sub-parameters.
     * @param context A parsed JSON-LD context.
     * @param externalContextsCallback Callback for external contexts.
     * @param classReference Class reference of the class component owning this parameter.
     * @param parameterData Parameter data with nested range.
     * @param parameters The array of parameters of the owning class, which will be appended to.
     * @param subParamData The sub-parameter of the parameter with nested range.
     * @param fieldId The @id of the field.
     * @param scope The current field scope.
     */
    constructFieldDefinitionNested(context: JsonLdContextNormalized, externalContextsCallback: ExternalContextCallback, classReference: ClassLoaded, parameterData: ParameterData<ParameterRangeResolved> & {
        range: {
            type: 'nested';
        };
    }, parameters: ParameterDefinition[], subParamData: ParameterData<ParameterRangeResolved>, fieldId: string, scope: FieldScope): Promise<ConstructorFieldDefinition>;
    /**
     * Construct a parameter definition from the given parameter data with raw range.
     * @param context A parsed JSON-LD context.
     * @param classReference Class reference of the class component owning this parameter.
     * @param parameterData Parameter data.
     * @param range Range of this parameter data.
     * @param fieldId The @id of the field.
     */
    constructParameterRaw(context: JsonLdContextNormalized, classReference: ClassLoaded, parameterData: ParameterData<ParameterRangeResolved>, range: string, fieldId: string): ParameterDefinition;
    /**
     * Construct a parameter definition from the given parameter data with an undefined range.
     * @param context A parsed JSON-LD context.
     * @param classReference Class reference of the class component owning this parameter.
     * @param parameterData Parameter data.
     * @param fieldId The @id of the field.
     */
    constructParameterRangeUndefined(context: JsonLdContextNormalized, classReference: ClassLoaded, parameterData: ParameterData<ParameterRangeResolved>, fieldId: string): ParameterDefinition;
    /**
     * Construct a parameter definition from the given parameter data with class reference range.
     * @param context A parsed JSON-LD context.
     * @param externalContextsCallback Callback for external contexts.
     * @param classReference Class reference of the class component owning this parameter.
     * @param parameterData Parameter data.
     * @param range Range of this parameter data.
     * @param fieldId The @id of the field.
     */
    constructParameterClass(context: JsonLdContextNormalized, externalContextsCallback: ExternalContextCallback, classReference: ClassReference, parameterData: ParameterData<ParameterRangeResolved>, range: ClassReference, fieldId: string): Promise<ParameterDefinition>;
    /**
     * Fill in the optional parameter definition values based on the given parameter data.
     * @param parameterDefinition A paramater definition to fill in.
     * @param parameterData Parameter data to read from.
     */
    populateOptionalParameterFields(parameterDefinition: ParameterDefinition, parameterData: ParameterData<ParameterRangeResolved>): void;
}
export interface ComponentConstructorArgs {
    packageMetadata: PackageMetadata;
    contextConstructor: ContextConstructor;
    pathDestination: PathDestinationDefinition;
    classAndInterfaceIndex: ClassIndex<ClassReferenceLoaded>;
    classConstructors: ClassIndex<ConstructorData<ParameterRangeResolved>>;
    externalComponents: ExternalComponents;
    contextParser: ContextParser;
}
export interface PathDestinationDefinition {
    /**
     * Absolute path to the package root.
     */
    packageRootDirectory: string;
    /**
     * Absolute path to the package source directory.
     */
    originalPath: string;
    /**
     * Absolute path to the package components target directory.
     */
    replacementPath: string;
}
export interface FieldScope {
    /**
     * All parent field names for the current scope.
     */
    parentFieldNames: string[];
    /**
     * A hash containing all previously created field names, to ensure uniqueness.
     */
    fieldIdsHash: Record<string, number>;
}
export declare type ExternalContextCallback = (contextUrl: string) => void;
