"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ContextConstructor = void 0;
const componentsjs_1 = require("componentsjs");
const semverMajor = require("semver/functions/major");
/**
 * Constructs a JSON-LD context for a given package..
 */
class ContextConstructor {
    constructor(args) {
        this.packageMetadata = args.packageMetadata;
        this.typeScopedContexts = args.typeScopedContexts;
    }
    /**
     * Determine a JSON-LD prefix for the given package name.
     * It will take the first letters of the name's components.
     * @param packageName A package name.
     */
    static getPackageNamePrefix(packageName) {
        return packageName
            .replace(/@/gu, '')
            .split(/[/-]/u)
            .map(part => part.charAt(0))
            .join('');
    }
    /**
     * Construct a JSON-LD context.
     * @param components Optional component definitions to provide shortcuts for.
     */
    constructContext(components) {
        var _a;
        // Determine a compact prefix to represent this package.
        const prefix = (_a = this.packageMetadata.prefix) !== null && _a !== void 0 ? _a : ContextConstructor.getPackageNamePrefix(this.packageMetadata.name);
        // Determine component shortcuts if provided.
        const componentShortcuts = components ? this.constructComponentShortcuts(components) : {};
        return {
            '@context': [
                componentsjs_1.PrefetchedDocumentLoader.CONTEXT_URL,
                Object.assign({ npmd: 'https://linkedsoftwaredependencies.org/bundles/npm/', [prefix]: `npmd:${this.packageMetadata.name}/`, [`files-${prefix}`]: `${prefix}:^${semverMajor(this.packageMetadata.version)}.0.0/` }, componentShortcuts),
            ],
        };
    }
    /**
     * Construct a hash of component shortcuts.
     * @param componentDefinitions Component definitions.
     */
    constructComponentShortcuts(componentDefinitions) {
        const shortcuts = {};
        for (const componentDefinition of Object.values(componentDefinitions)) {
            for (const component of componentDefinition.components) {
                // Shortcut name for a class contains no special characters
                // Regex will always match
                const match = (/[a-z0-9]*$/iu.exec(component['@id']));
                shortcuts[match[0]] = {
                    '@id': component['@id'],
                    '@prefix': true,
                };
                // Generate type-scoped context when enabled
                if (this.typeScopedContexts) {
                    const typeScopedContext = {};
                    for (const parameter of component.parameters) {
                        typeScopedContext[parameter['@id'].slice(Math.max(0, component['@id'].length + 1))] = Object.assign(Object.assign({ '@id': parameter['@id'] }, parameter.range === 'rdf:JSON' ? { '@type': '@json' } : {}), parameter.unique || parameter.range === 'rdf:JSON' ? {} : { '@container': '@list' });
                    }
                    shortcuts[match[0]]['@context'] = typeScopedContext;
                }
            }
        }
        return shortcuts;
    }
}
exports.ContextConstructor = ContextConstructor;
//# sourceMappingURL=ContextConstructor.js.map